# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only

"""This module provides utilities to read the grm grammar format"""

from pypargen.grm.parser import GrmParser

__all__ = ["GrmParser"]



# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only

import pytest
import io
from pypargen.base import parser


def test_base_grammar():
    grammar = parser.BaseGrammar([('a', ['a', 'b']), ('a', []),
                                  ('b', ['"b"'])])
    parsr = parser.BaseParser(grammar)
    assert parsr.grammar == grammar


@pytest.mark.xfail(raises=NotImplementedError)
def test_base_parse_err():
    grammar = parser.BaseGrammar([('a', ['a', 'b']), ('a', []),
                                  ('b', ['"b"'])])
    parsr = parser.BaseParser(grammar)
    inputstr = "aabb"
    inputbuf = io.StringIO(inputstr)
    parsr.parse(inputbuf)



# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only

import pytest
import io
from pypargen.base import lexer


def test_base_terminals():
    terminals = ['"a"', '"b"']
    inpt = "aabb"
    inputbuf = io.StringIO(inpt)
    lex = lexer.BaseLexer(terminals, inputbuf)
    assert lex.terminals == terminals


@pytest.mark.xfail(raises=NotImplementedError)
def test_base_iter_err():
    terminals = ['"a"', '"b"']
    inpt = "aabb"
    inputbuf = io.StringIO(inpt)
    lex = lexer.BaseLexer(terminals, inputbuf)
    list(lex)



# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only

from pypargen.base import token


def test_token_nonterm():
    tok = token.Token("terminal", "hello")
    assert str(tok) == "terminal(hello)"
    assert repr(tok) == "<terminal(hello)>"
    assert tok


def test_token_term():
    tok = token.Token('"[1-9][0-9]*"', "23")
    assert str(tok) == '"[1-9][0-9]*"("23")'
    assert repr(tok) == '<"[1-9][0-9]*"("23")>'
    assert tok


def test_token_empty():
    tok = token.Token('', 'hello')
    assert not tok



# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only

"""This example shows how to build a parser that "counts"
the number of occurences of a and b"""

import pypargen as pgen
import sys

g = pgen.lr1.Grammar()
g.append(pgen.Rule('c', ['a', 'b', '"\n\n*"']))
g.append(pgen.Rule('a', ['a', '"a"']))
g.append(pgen.Rule('a', []))
g.append(pgen.Rule('b', ['b', '"b"']))
g.append(pgen.Rule('b', ['"b"']))

# Optional here, (By default, it takes the first rule's LHS as start)
g.start = 'c'


def root(a, b, _):
    return [a, b]


def a_inc(a, _):
    return f"{int(a[:-1]) + 1}a"


def a_init():
    return "0a"


def b_inc(b, _):
    return f"{int(b[:-1]) + 1}b"


def b_init(c):
    return '1b'


callbacks = [root, a_inc, a_init, b_inc, b_init]
parser = pgen.lr1.Parser(g, callbacks, sys.stdin)
counts = parser.parse()

print(counts)



# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only

from typing import NamedTuple


class Rule(NamedTuple):
    """Rule is a representation of a grammar rule."""
    lhs: str
    rhs: list[str]

    def __hash__(self) -> int:
        return hash((self.lhs, tuple(self.rhs)))

    def __str__(self) -> str:
        return "\t-> ".join([self.lhs, ' '.join(self.rhs) if self.rhs else 'ϵ'])

    def __repr__(self) -> str:
        return f"<{self.__str__()}>"



# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only

from typing import NamedTuple


class Token(NamedTuple):
    """Token is a representation of a grammar token"""
    type: str
    content: any

    def __str__(self):
        if self.type.startswith('"'):
            return f"{self.type}(\"{self.content}\")"
        return f"{self.type}({self.content})"

    def __repr__(self):
        return f"<{self.__str__()}>"

    def __bool__(self):
        return bool(self.type)



# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only

import pytest
from pypargen.lr1 import grammar


@pytest.fixture
def item():
    return grammar.Item('a', ['b', 'c'], 0, '"a"')


def test_item_str_repr(item: grammar.Item):
    assert str(item) == '[a -> . b c, "a"]'
    assert repr(item) == '<[a -> . b c, "a"]>'


def test_item_copy(item: grammar.Item):
    item2 = item.copy()
    assert item.lhs == item2.lhs
    assert item.rhs == item2.rhs
    assert item.pos == item2.pos
    assert item.lookahead == item2.lookahead
    assert hash(item) == hash(item2)
    assert item == item2

    assert not item.done
    assert not item2.done
    item2.pos += 2
    assert item2.done


@pytest.fixture
def palindrome():
    return grammar.Grammar([('S', ['"a"', 'S', '"a"']),
                            ('S', ['"b"', 'S', '"b"']), ('S', ['"c"'])])


def test_closure_palindrome(palindrome: grammar.Grammar):
    items = [grammar.Item('S', ['"a"', 'S', '"a"'], 0, '"a"')]
    closure = palindrome.closure(items)
    assert closure == items

    items = [grammar.Item('S', ['"a"', 'S', '"a"'], 1, '"a"')]
    closure = palindrome.closure(items)
    true_closure = items.copy()
    true_closure.append(grammar.Item('S', ['"a"', 'S', '"a"'], 0, '"a"'))
    true_closure.append(grammar.Item('S', ['"b"', 'S', '"b"'], 0, '"a"'))
    true_closure.append(grammar.Item('S', ['"c"'], 0, '"a"'))
    assert closure == true_closure

    items = [grammar.Item('S', ['"a"', 'S', '"a"'], 2, '"a"')]
    closure = palindrome.closure(items)
    assert closure == items

    items = [grammar.Item('S', ['"a"', 'S', '"a"'], 3, '"a"')]
    closure = palindrome.closure(items)
    assert closure == items


def test_goto_palindrome(palindrome: grammar.Grammar):
    items = [grammar.Item('S', ['"a"', 'S', '"a"'], 0, '"a"')]
    goto = palindrome.goto(items, '"a"')
    true_goto = [grammar.Item('S', ['"a"', 'S', '"a"'], 1, '"a"')]
    true_goto = palindrome.closure(true_goto)
    assert goto == true_goto

    items = [grammar.Item('S', ['"a"', 'S', '"a"'], 1, '"a"')]
    goto = palindrome.goto(items, 'S')
    true_goto = [grammar.Item('S', ['"a"', 'S', '"a"'], 2, '"a"')]
    assert goto == true_goto

    items = [grammar.Item('S', ['"a"', 'S', '"a"'], 2, '"a"')]
    goto = palindrome.goto(items, '"a"')
    true_goto = [grammar.Item('S', ['"a"', 'S', '"a"'], 3, '"a"')]
    assert goto == true_goto

    items = [grammar.Item('S', ['"a"', 'S', '"a"'], 3, '"a"')]
    goto = palindrome.goto(items, '"a"')
    true_goto = []
    assert goto == true_goto


def test_table_palindrome(palindrome: grammar.Grammar):
    # TODO: Check the parse table
    palindrome.parse_table()


@pytest.mark.xfail(strict=True, raises=grammar.ReduceReduceConflict)
def test_table_lr2():
    rules = [("root", ['a1', 'b', '"x"']), ("root", ['a2', 'b', '"y"']),
             ("a1", ['"a"']), ("a2", ['"a"']), ("b", ['"b"'])]
    grm = grammar.Grammar(rules, "root")
    grm.parse_table()


@pytest.mark.xfail(strict=True, raises=grammar.ShiftReduceConflict)
def test_table_pda():
    palindrome = grammar.Grammar([('S', ['"a"', 'S', '"a"']),
                                  ('S', ['"b"', 'S', '"b"']), ('S', [])])
    palindrome.parse_table()


def test_eps_grammar():
    g = grammar.Grammar([('a', ['b', 'c']), ('b', []), ('b', ['"b"']),
                         ('c', ['c', '"c"']), ('c', ['"c"'])])
    # TODO: Check parse table
    g.parse_table()



# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only
"""This example shows how to build a json parser"""

import pypargen as pgen
import sys
import pathlib

grm_parser = pgen.GrmParser(pgen.Lexer)
with (pathlib.Path(__file__).parent / "json.grm").open() as fp:
    json_grm = grm_parser.parse(fp)


def nop(a):
    return a


callbacks = [nop] * 4


def true(_):
    return True


def false(_):
    return False


def null(_):
    return None


callbacks += [true, false, null]


def kvpair(key, _colon, val):
    return (key, val)


def kvpairs_append(kvpairs, _comma, kvpair):
    key, val = kvpair
    kvpairs[key] = val
    return kvpairs


def kvpairs_init(kvpair):
    key, val = kvpair
    return {key: val}


def object_empty(_left, _right):
    return {}


def object_filled(_left, kvpairs, _right):
    return kvpairs


callbacks += [
    kvpair, kvpairs_append, kvpairs_init, object_empty, object_filled
]


def vals_append(vals, _comma, val):
    vals.append(val)
    return vals


def vals_init(val):
    return [val]


def array_empty(_left, _right):
    return []


def array_filled(_left, vals, _right):
    return vals


callbacks += [vals_append, vals_init, array_empty, array_filled]


def hx(hstr):
    return chr(int(hstr, 16))


def esc(_bks, echr):
    if echr not in "bfnrt":
        return echr
    if echr == 'b':
        return '\b'
    if echr == 'f':
        return '\f'
    if echr == 'n':
        return '\n'
    if echr == 'r':
        return '\r'
    if echr == 't':
        return '\t'


def hxesc(_bks, _u, hx):
    return hx


def str_join(*args):
    return ''.join(args)


def string(_left, string1, _right):
    return string1


callbacks += [hx, esc, hxesc, nop]
callbacks += [str_join] * 2
callbacks += [string]

# Number
callbacks += [float]

json_parser = pgen.Parser(json_grm,
                          callbacks,
                          pgen.Lexer,
                          whitespaces=" \t\r\n")

parsed = json_parser.parse(sys.stdin)
print(parsed)



# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only

"""A parser generator built with Python"""

__version__ = "0.3.0"

from pypargen.base import *
from pypargen.lexer import *
from pypargen.lr1 import *
from pypargen.grm import *



# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only

from pypargen.base.rule import Rule


class BaseGrammar(list[Rule]):
    """Grammar defines a context free grammar. It is simply a list of rules."""

    def __init__(self, iterable=(), start=None):
        """Create grammar with iterable of rules.
        A rule can either be a tuple of lhs (str) and rhs (list of str)
        or the Rule object."""
        super().__init__([Rule(*x) for x in iterable])
        assert "__root__" not in self.nonterminals, \
            "__root__ is a reserved nonterminal"
        if start:
            assert start in self.nonterminals,\
                    "Start symbol must be valid nonterminal"
        self._start = start
        self._firsts = {('$', ): {'$'}, (): {'ϵ'}}

    @property
    def start(self) -> str:
        """Returns the start symbol for the grammar."""
        if self._start:
            return self._start
        assert len(self) >= 1, "No rules added to the grammar yet"
        return self[0].lhs

    @start.setter
    def start(self, start: str):
        """Sets the start symbol for the grammar. Checks for validity"""
        assert start in self.nonterminals,\
            "Start symbol must be a valid nonterminal"
        self._start = start

    @property
    def terminals(self) -> list[str]:
        """Returns all the terminals found from the grammar"""
        terms = []
        for _, rhs in self:
            for tok in rhs:
                if tok.startswith('"') and tok not in terms:
                    terms.append(tok)
        return terms

    @property
    def nonterminals(self) -> list[str]:
        """Returns all the nonterminals from the grammar"""
        nonterms = []
        for lhs, _ in self:
            if lhs not in nonterms:
                nonterms.append(lhs)
        return nonterms

    def __str__(self) -> str:
        return '\n'.join(map(str, self)) + '\n'

    def first(self, tokens: list[str]) -> list[str]:
        """Returns the first set for a list of tokens"""
        if (ttokens := tuple(tokens)) in self._firsts:
            return self._firsts[ttokens]

        if tokens[0].startswith('"'):
            firsts = [tokens[0]]
            # Memoization
            self._firsts[tuple(tokens)] = firsts
            return firsts

        firsts = []
        for lhs, rhs in self:
            if tokens[0] == lhs:
                recursive = False
                # NOTE: Only first RHS token verified
                if rhs and lhs == rhs[0]:
                    # Left recursive, check if empty
                    for lhsi, rhsi in self:
                        if lhsi == lhs and len(rhsi) <= 0:
                            recursive = True
                            break
                    if not recursive:
                        # Recursive without epsilon rule
                        continue

                for f in self.first(rhs[1:] if recursive else rhs):
                    if f not in firsts:
                        firsts.append(f)

        if 'ϵ' in firsts:
            firsts.remove('ϵ')
            for f in self.first(tokens[1:]):
                if f not in firsts:
                    firsts.append(f)

        # Memoization
        self._firsts[tuple(tokens)] = firsts

        return firsts



# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only

import pytest
from pypargen.base import grammar


@pytest.mark.xfail(strict=True, raises=AssertionError)
def test_empty_grammar():
    empty = grammar.BaseGrammar()
    empty.start


def test_palindrome():
    palindrome = grammar.BaseGrammar([('S', ['"a"', 'S', '"a"']),
                                      ('S', ['"b"', 'S', '"b"']), ('S', [])])
    assert palindrome.start == 'S'
    assert palindrome.terminals == ['"a"', '"b"']
    assert palindrome.nonterminals == ['S']
    assert palindrome.first(['S']) == ['"a"', '"b"', 'ϵ']
    assert str(palindrome) == """S\t-> "a" S "a"
S\t-> "b" S "b"
S\t-> ϵ
"""


def test_math():
    math_rules = [("atom", ['"[1-9][0-9]*"']),
                  ("atom", [r'"\("', "sub", r'"\)"']),
                  ("div", ["div", '"/"', "atom"]), ("div", ["atom"]),
                  ("mul", ["mul", r'"\*"', "div"]), ("mul", ["div"]),
                  ("add", ["add", r'"\+"', "mul"]), ("add", ["mul"]),
                  ("sub", ["sub", '"-"', "add"]), ("sub", ["add"])]
    math = grammar.BaseGrammar(math_rules, 'sub')
    assert math.start == 'sub'
    assert math.terminals == [
        '"[1-9][0-9]*"', r'"\("', r'"\)"', '"/"', r'"\*"', r'"\+"', '"-"'
        ]
    assert math.nonterminals == ["atom", "div", "mul", "add", "sub"]
    assert math.first(["sub"]) == ['"[1-9][0-9]*"', r'"\("']
    assert math.first(["add"]) == ['"[1-9][0-9]*"', r'"\("']
    assert math.first(["mul"]) == ['"[1-9][0-9]*"', r'"\("']
    assert math.first(["div"]) == ['"[1-9][0-9]*"', r'"\("']
    assert math.first(["atom"]) == ['"[1-9][0-9]*"', r'"\("']


def test_eps_first():
    rules = [('a', []), ('b', []), ('c', ['a', 'b', '"a"'])]
    grm = grammar.BaseGrammar(rules)
    grm.start = 'c'

    assert grm.first(['a']) == ['ϵ']
    assert grm.first(['b']) == ['ϵ']
    assert grm.first(['c']) == ['"a"']


def test_eps_end():
    rules = [('a', []), ('a', ['"a"'])]
    grm = grammar.BaseGrammar(rules)

    assert grm.start == 'a'
    assert grm.first(['a', '$']) == ['"a"', '$']


def test_eps_recursive():
    rules = [('a', ['b', '"a"']), ('b', ['b', '"b"']), ('b', [])]
    grm = grammar.BaseGrammar(rules)

    assert grm.start == 'a'
    assert grm.first(['a', '$']) == ['"b"', '"a"']


@pytest.mark.xfail(strict=True, raises=AssertionError)
def test_reserved():
    grammar.BaseGrammar([('__root__', ['"a"', 'S', '"a"']),
                         ('__root__', ['"b"', 'S', '"b"']),
                         ('__root__', ['"c"'])])


@pytest.mark.xfail(strict=True, raises=AssertionError)
def test_invalid_start():
    grammar.BaseGrammar([('S', ['"a"', 'S', '"a"']),
                         ('S', ['"b"', 'S', '"b"']), ('S', [])], 'T')


@pytest.mark.xfail(strict=True, raises=AssertionError)
def test_invalid_start_set():
    g = grammar.BaseGrammar([('S', ['"a"', 'S', '"a"']),
                             ('S', ['"b"', 'S', '"b"']), ('S', [])])
    g.start = 'T'



# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only

from typing import Union

from pypargen.base.grammar import BaseGrammar
from pypargen.base.rule import Rule


class Item:
    """Item is an LR(1) item"""

    def __init__(self, lhs: str, rhs: list[str], pos: int, lookahead: str):
        """Initialize the grammar item"""
        assert 0 <= pos <= len(rhs), "Dot position out of range"
        self.lhs = lhs
        self.rhs = rhs
        self.pos = pos
        self.lookahead = lookahead

    @property
    def done(self) -> bool:
        """True if the position is at the end"""
        return self.pos >= len(self.rhs)

    def __eq__(self, other: "Item") -> bool:
        return self.lhs == other.lhs and self.rhs == other.rhs and \
                self.pos == other.pos and self.lookahead == other.lookahead

    def __hash__(self) -> int:
        return hash((self.lhs, tuple(self.rhs), self.pos, self.lookahead))

    def __str__(self) -> str:
        rhs = self.rhs.copy()
        rhs.insert(self.pos, '.')
        return f"[{self.lhs} -> {' '.join(rhs)}, {self.lookahead}]"

    def __repr__(self) -> str:
        return f"<{self.__str__()}>"

    def copy(self) -> "Item":
        return Item(self.lhs, self.rhs.copy(), self.pos, self.lookahead)


class ShiftReduceConflict(Exception):
    """Exception thrown when Shift/Reduce conflict is found"""

    def __init__(self, grm: "Grammar", items: set[Item], lookahead: str):
        """Initialize exception with conflicting items and lookahead"""
        conflicts = [item for item in items if grm.goto([item], lookahead)]
        conflicts += [
            item for item in items if item.done and item.lookahead == lookahead
        ]
        conflicts = set(conflicts)
        msg = '\n'.join(map(str, conflicts))
        super().__init__(f"Shift/Reduce Conflict:\n{msg}\nlookahead: {lookahead}")


class ReduceReduceConflict(Exception):
    """Exception thrown when Reduce/Reduce conflict is found"""

    def __init__(self, rule1: Rule, rule2: Rule):
        """Initialize exception with conflicting reduction rules"""
        msg = '\n'.join(map(str, [rule1, rule2]))
        super().__init__(f"Reduce/Reduce Conflict:\n{msg}")


class Grammar(BaseGrammar):
    """Grammar is a LR(1) grammar. The parse_table method gives the parsing
    table for the grammar."""

    def closure(self, items: list[Item]) -> list[Item]:
        """Closure calculates the closure a for set of items."""
        assert len(items) == len(set(items)), "Items must not be repeated"

        closure_items = {k: None for k in items}
        new_items = {}
        while True:
            for item in closure_items:
                if item.done:
                    continue
                if item.rhs[item.pos].startswith('"'):
                    continue
                for lhs, rhs in self:
                    if item.rhs[item.pos] == lhs:
                        for lookahead in self.first(item.rhs[item.pos + 1:] +
                                                    [item.lookahead]):
                            if lookahead != 'ϵ':
                                new_items[Item(lhs, rhs, 0, lookahead)] = None
            if set(closure_items).issuperset(new_items):
                break
            closure_items |= new_items
            new_items = {}
        return list(closure_items)

    def goto(self, items: list[Item], token: str) -> list[Item]:
        """Goto calculates the set of items to go to when a token is seen."""
        assert len(items) == len(set(items)), "Items must not be repeated"

        goto = {}
        for item in items:
            if item.done:
                continue

            if item.rhs[item.pos] == token:
                gitem = item.copy()
                gitem.pos += 1
                goto[gitem] = None
        return self.closure(goto)

    def parse_table(self) -> list[dict[str, Union[int, str]]]:
        """parse_table gives the parsing table for the grammar."""
        init_item = Item("__root__", [self.start], 0, '$')
        set_of_items = [self.closure([init_item])]

        table = [{}]
        symbols = self.terminals
        symbols.extend(self.nonterminals)

        # Dragon book: 4.7.1 Canonical LR(1) Parser
        # Build goto table
        while True:
            added = False
            for idx, items in enumerate(set_of_items):
                for sym in symbols:
                    if gitems := self.goto(items, sym):
                        if gitems not in set_of_items:
                            set_of_items.append(gitems)
                            table.append({})
                            added = True
                        table[idx][sym] = set_of_items.index(gitems)
                if added:
                    break
            if not added:
                break

        # Fill the reduction entries
        for idx, items in enumerate(set_of_items):
            for item in items:
                if item.done:

                    # If conflict, raise proper error
                    if conflict := table[idx].get(item.lookahead, None):
                        if isinstance(conflict, int):
                            raise ShiftReduceConflict(self, items,
                                                      item.lookahead)

                        rule1 = self[int(conflict[1:])]
                        rule2 = Rule(item.lhs, item.rhs)
                        raise ReduceReduceConflict(rule1, rule2)

                    if item.lhs == "__root__":
                        table[idx][item.lookahead] = 'c'
                        continue
                    table[idx][item.lookahead] = \
                        f"r{self.index((item.lhs, item.rhs))}"

        return table


__all__ = ["Grammar"]



# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only

import pytest
import io
import re
from pypargen.lexer import pyre, lexer


@pytest.mark.parametrize("lexerClass", [pyre.PyRELexer, lexer.Lexer])
def test_whitespaces(lexerClass):
    terminals = ['"a"', '"b"']
    input = "	a a b 	b"
    whitespaces = " \t"
    inputbuf = io.StringIO(input)
    lexer1 = lexerClass(terminals, inputbuf, whitespaces=whitespaces)
    assert lexer1.terminals == terminals

    tokens = list(lexer1)

    assert ''.join(map(lambda x: x.content,
                       tokens[:-1])) == re.sub(f"[{whitespaces}]", "", input)

    true_token_types = [0, 0, 1, 1]
    assert [x.type for x in tokens
            ] == [terminals[i] for i in true_token_types] + ['$']


@pytest.mark.parametrize("lexerClass", [pyre.PyRELexer, lexer.Lexer])
def test_palindrome(lexerClass):
    terminals = ['"a"', '"b"']
    input = "aabb"
    inputbuf = io.StringIO(input)
    lexer1 = lexerClass(terminals, inputbuf)
    assert lexer1.terminals == terminals

    tokens = list(lexer1)

    assert ''.join(map(lambda x: x.content, tokens[:-1])) == input

    true_token_types = [0, 0, 1, 1]
    assert [x.type for x in tokens
            ] == [terminals[i] for i in true_token_types] + ['$']


@pytest.mark.parametrize("lexerClass", [pyre.PyRELexer, lexer.Lexer])
def test_math(lexerClass):
    terminals = [
        '"[1-9][0-9]*"', r'"\("', r'"\)"', '"/"', r'"\*"', r'"\+"', '"-"'
    ]
    input = "(1+2)/(4-1)"
    inputbuf = io.BytesIO(input.encode())

    if lexerClass == lexer.Lexer:
        terminals[5] = '"+"'
    lexer1 = lexerClass(terminals, inputbuf)
    assert lexer1.terminals == terminals

    tokens = list(lexer1)

    assert ''.join(map(lambda x: x.content, tokens[:-1])) == input

    true_token_types = [1, 0, 5, 0, 2, 3, 1, 0, 6, 0, 2]
    assert [x.type for x in tokens
            ] == [terminals[i] for i in true_token_types] + ['$']


@pytest.mark.parametrize("lexerClass", [pyre.PyRELexer, lexer.Lexer])
@pytest.mark.xfail(strict=True, raises=pyre.UnexpectedCharacter)
def test_invalid(lexerClass):
    terminals = ['"a"', '"b"']
    input = "aabxab"
    inputbuf = io.StringIO(input)
    lexer1 = lexerClass(terminals, inputbuf)
    assert lexer1.terminals == terminals

    list(lexer1)


@pytest.mark.parametrize("lexerClass", [pyre.PyRELexer, lexer.Lexer])
def test_active(lexerClass):
    terminals = ['"[a-z]"', '"[A-Za-z]"']
    inputstr = "abcAbc"
    inputbuf = io.StringIO(inputstr)
    lexer1 = lexerClass(terminals, inputbuf)
    i = 0

    while (tok := lexer1.nextToken(terminals)).type != '$':
        if i == 3:
            terminals = terminals[1:]
        assert tok.type == terminals[0]
        assert tok.content == inputstr[i]
        i += 1


@pytest.mark.parametrize("lexerClass", [pyre.PyRELexer, lexer.Lexer])
@pytest.mark.xfail(strict=True, raises=pyre.UnregisteredTerminal)
def test_active_invalid(lexerClass):
    terminals = ['"[a-z]"', '"[A-Za-z]"']
    inputstr = "abcABC"
    inputbuf = io.StringIO(inputstr)
    lexer1 = lexerClass(terminals, inputbuf)
    i = 0

    while (tok := lexer1.nextToken(terminals)).type != '$':
        assert tok.type == terminals[0]
        assert tok.content == inputstr[i]
        i += 1
        if i == 3:
            terminals = ['"[A-Z]"']
    raise RuntimeError("Should not reach here")



# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only

import pytest
from pypargen.lexer import fsm


def test_nfa_node_id():
    a = fsm.NFANode()
    b = fsm.NFANode()
    c = fsm.NFANode()
    d = fsm.NFANode()
    assert a != b != c != d
    assert a == a
    assert b == b
    assert c == c
    assert d == d
    assert hash(a) != hash(b) != hash(c) != hash(d)


def test_nfa_node_transition():
    start = fsm.NFANode()
    start.add_transition('a', fsm.NFANode('atok'))
    start.add_transition('b', fsm.NFANode())
    start.add_transition('c', fsm.NFANode())
    start.add_transitions('c', [fsm.NFANode(), fsm.NFANode(), fsm.NFANode()])
    assert list(start.keys()) == list("abc")
    assert start not in start['a']
    assert start not in start['b']
    assert start not in start['c']
    assert len(start['c']) == 4
    a = list(start['a'])[0]
    assert a.token == 'atok'
    assert not list(start['b'])[0].token


def test_nfa_node_multi_chars_transition():
    start = fsm.NFANode()
    start.add_chars_transition('hello', fsm.NFANode('got'))
    assert set(start.keys()) == set('hello')
    assert start['h'] == start['e'] == start['l'] == start['o']


@pytest.mark.xfail(strict=True, raises=AssertionError)
def test_nfa_node_transition_invalid():
    start = fsm.NFANode()
    start.add_transition('hello', fsm.NFANode())


def test_nfa():
    nfa = fsm.NFA()
    nfa.start.add_transition('a', nfa.end)
    nfa.start.add_transition('b', nfa.end)
    nfa.start.add_transition('c', nfa.end)
    nfa.start.add_transition('c', nfa.end)
    nfa.start.add_transition('c', nfa.end)

    assert list(nfa.start.keys()) == list("abc")
    assert nfa.end in nfa.start['a']
    assert nfa.end in nfa.start['b']
    assert nfa.end in nfa.start['c']
    assert len(nfa.start['c']) == 1


def test_dfa_node():
    a = fsm.NFANode()
    b = fsm.NFANode()
    c = fsm.NFANode()
    d = fsm.NFANode('atok')
    a.add_transition('', b)
    b.add_transition('', c)
    c.add_transition('a', d)

    dfanode = fsm.DFANode({a})
    assert b in dfanode
    assert c in dfanode
    assert d not in dfanode
    assert d in dfanode.move('a')
    assert len(dfanode.move('a')) == 1
    assert not dfanode.tokens
    assert dfanode.move('a').tokens == {"atok"}


def test_dfa_node_multi_tokens():
    a = fsm.NFANode()
    b = fsm.NFANode('b')
    c = fsm.NFANode('c')
    a.add_transition('a', b)
    a.add_transition('a', c)

    dfanode = fsm.DFANode({a})
    assert dfanode.move('a').tokens == {'b', 'c'}


def test_dfa():
    nfa = fsm.NFA(end=fsm.NFANode("end"))
    a = fsm.NFANode()
    b = fsm.NFANode()
    c = fsm.NFANode()
    d = fsm.NFANode()

    a.add_transition('', b)
    b.add_transition('', c)
    c.add_transition('a', d)

    nfa.start.add_transition('a', a)
    d.add_transition('d', nfa.end)
    nfa.end.add_transition('a', d)

    dfa = fsm.DFA(nfa)
    assert dfa.match("aad") == ({"end"}, 3)
    assert dfa.match("aadaddd") == ({"end"}, 5)
    assert not dfa.match("aaa")
    assert not dfa.match("a")


def test_dfa_empty():
    nfa = fsm.NFA()
    nfa.start.add_transition('', nfa.end)
    nfa.end.token = "empty"

    dfa = fsm.DFA(nfa)

    assert dfa.match("hello")[1] == 0
    assert dfa.match("")[1] == 0
    assert dfa.match("abcd")[1] == 0


def test_combine_dfa():
    hello = fsm.NFA()
    h = fsm.NFANode()
    e = fsm.NFANode()
    l1 = fsm.NFANode()
    l2 = fsm.NFANode()

    hello.start.add_transition('h', h)
    h.add_transition('e', e)
    e.add_transition('l', l1)
    l1.add_transition('l', l2)
    l2.add_transition('o', hello.end)
    hello.end.token = "hello"

    hola = fsm.NFA()
    h = fsm.NFANode()
    o = fsm.NFANode()
    l1 = fsm.NFANode()

    hola.start.add_transition('h', h)
    h.add_transition('o', o)
    o.add_transition('l', l1)
    l1.add_transition('a', hola.end)
    hola.end.token = "hola"

    hellohola = fsm.NFA()
    hellohola.start.add_transitions('', [hello.start, hola.start])

    hhDFA = fsm.DFA(hellohola)

    assert hhDFA.match('holahello') == ({'hola'}, 4)
    assert hhDFA.match('hellohola') == ({'hello'}, 5)
    assert not hhDFA.match('helo')



# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only

import io
from typing import Callable, Optional

from pypargen.base.lexer import BaseLexer
from pypargen.lexer.pyre import PyRELexer
from pypargen.base.parser import BaseParser
from pypargen.base.token import Token
from pypargen.lr1.grammar import Grammar


class Parser(BaseParser):
    """Parser is an LR(1) parser"""

    def __init__(self,
                 grammar: Grammar,
                 callbacks: list[Callable],
                 lexerClass: type[BaseLexer] = PyRELexer,
                 whitespaces: Optional[str] = None):
        """Initialize parser with LR(1) grammar, callbacks and input stream

        callbacks is a list of functions that corresponding to the rules.
        For example, if the rules are:
        S -> a S a
        S -> b S b
        S -> c

        The callbacks could be:
        def rule1(a, S, _):
            return a+S

        def rule2(b, S, _):
            return b+S

        def rule3(_):
            return ""

        callbacks = [rule1, rule2, rule3]

        Note that the callbacks should take the same number of arguments as RHS
        and return a single value that will be used for next callback.
        It could be a parse (sub)tree, calculated expression etc."""
        assert len(grammar) == len(callbacks),\
            "Callbacks and grammar must be of same size"
        super().__init__(grammar, lexerClass, whitespaces)
        self.table = grammar.parse_table()
        self.callbacks = callbacks

    def parse(self, inpt: io.RawIOBase) -> any:
        """Start parsing the input stream and provide the final result from\
        callbacks."""
        lexer = self.lexerClass(self.grammar.terminals, inpt, self.whitespaces)
        states = [0]
        tokens = [None]

        token = lexer.nextToken(
            [x for x in self.table[0] if x.startswith('"')])
        while True:
            if (nxt := self.table[states[-1]].get(token.type, None)) is None:
                if token.type == '$':
                    raise EOFError("Unexpected EOF")
                raise SyntaxError("Unexpected token",
                                  ("input", 0, 0, token.type))
            if isinstance(nxt, int):
                states.append(nxt)
                tokens.append(token)

                # Read the next token
                token = lexer.nextToken(
                    [x for x in self.table[nxt] if x.startswith('"')])
                continue

            if nxt == 'c':
                assert len(states) == len(tokens) == 2
                return tokens[1].content

            # Get the reduction rule
            rule_num = int(nxt[1:])
            rule = self.grammar[rule_num]

            # Get RHS tokens and pop them off the stack
            rhs_len = -len(rule.rhs) if rule.rhs else len(tokens)
            rhs_tokens = tokens[rhs_len:]
            tokens = tokens[:rhs_len]
            states = states[:rhs_len]

            # Reduce RHS to LHS with callback
            lhs_content = self.callbacks[rule_num](
                *[x.content for x in rhs_tokens])
            lhs_token = Token(rule.lhs, lhs_content)
            tokens.append(lhs_token)

            # Goto
            nxt = self.table[states[-1]][rule.lhs]
            states.append(nxt)


__all__ = ["Parser"]



# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only

import io
from typing import Optional

from pypargen.base.lexer import BaseLexer, UnexpectedCharacter,\
        UnregisteredTerminal
from pypargen.base.token import Token
from pypargen.lexer import fsm, re


class Lexer(BaseLexer):
    """Lexer built with RE parser and FSM library right inside pypargen."""

    def __init__(self,
                 terminals: list[str],
                 inpt: io.RawIOBase,
                 whitespaces: Optional[str] = None):
        """Initiallize the lexer. Similar to base initialization arguments."""
        super().__init__(terminals, inpt, whitespaces)
        if self.whitespaces:
            self.whitespaces = list(self.whitespaces)
        else:
            self.whitespaces = []
        self._re_parser = re.REParser()
        self.nfa_starts = {}

        # Build the NFA's and combine them
        for term in self.terminals:
            nfa = self._re_parser.parse(term[1:-1])
            nfa.end.token = term
            self.nfa_starts[term] = nfa.start

        self.stopped = False
        self.pos = 0
        self.next_char()

    def next_char(self):
        self.buf = self.input.read(1)
        self.pos += len(self.buf)
        if hasattr(self.buf, 'decode'):
            self.buf = self.buf.decode()

    def nextToken(self, terminals: Optional[list[str]] = None) -> Token:
        """Request next token from the Lexer. Pass optional terminals to look
        for only these terminals."""
        if self.stopped:
            raise StopIteration

        # First, skip whitespaces
        while self.buf in self.whitespaces:
            self.next_char()

        if self.buf == '':
            self.stopped = True
            return Token('$', None)

        if terminals is None:
            terminals = self.terminals
        else:
            for term in terminals:
                if term not in self.terminals:
                    raise UnregisteredTerminal(term)

        state = fsm.DFANode({self.nfa_starts[term] for term in terminals})
        content = ''

        while self.buf != '':
            nstate = state.move(self.buf)
            if not nstate:
                break

            # Next state
            state = nstate
            content += self.buf

            # Read next
            self.next_char()

        if not state.tokens:
            raise UnexpectedCharacter(self.buf, self.pos, terminals)

        for term in terminals:
            if term in state.tokens:
                break
        else:
            raise UnexpectedCharacter(self.buf, self.pos, terminals)

        return Token(term, content)



# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only

import io

from pypargen.lr1 import Grammar, Parser
from pypargen.grm.grammar import grammar
from pypargen.lexer import PyRELexer
from pypargen.base.lexer import BaseLexer
from pypargen.base.rule import Rule


def join_str(*args):
    return ''.join(args)


# For all the terminals processing
callbacks = [join_str] * 26


def nop(a):
    return a


# For direct reductions
callbacks += [nop] * 3


def rhs_append(rhs, _ws, rhsc):
    rhs.append(rhsc)
    return rhs


def rhs_init(rhsc):
    return [rhsc]


# For building RHS
callbacks += [rhs_append, rhs_init]


def stmt(nont, _ws, _arrow, _ws2, rhs, _nl):
    return Rule(nont, rhs)


def stmt_eps(nont, _ws, _arrow, _ws2, _eps, _nl):
    return Rule(nont, [])


# For building statement
callbacks += [stmt, stmt_eps]


def grm_append(grm, stmt):
    grm.append(stmt)
    return grm


def grm_init():
    return Grammar()


# Final grammar!
callbacks += [grm_append, grm_init]


class GrmParser(Parser):
    def __init__(self, lexerClass: type[BaseLexer] = PyRELexer):
        super().__init__(grammar, callbacks, lexerClass)

    def parse(self, inpt: io.RawIOBase) -> Grammar:
        return super().parse(inpt)



# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only

import io
import pytest
from pypargen.lr1 import parser, grammar


@pytest.fixture
def math():
    math_rules = [("atom", ['"[1-9][0-9]*"']),
                  ("atom", [r'"\("', "sub", r'"\)"']),
                  ("div", ["div", '"/"', "atom"]), ("div", ["atom"]),
                  ("mul", ["mul", r'"\*"', "div"]), ("mul", ["div"]),
                  ("add", ["add", r'"\+"', "mul"]), ("add", ["mul"]),
                  ("sub", ["sub", '"-"', "add"]), ("sub", ["add"])]
    return grammar.Grammar(math_rules, "sub")


def test_math(math: grammar.Grammar):
    def convnum(a):
        return int(a)

    def brac(_, a, b):
        return a

    def div(a, _, b):
        return a / b

    def mul(a, _, b):
        return a * b

    def add(a, _, b):
        return a + b

    def sub(a, _, b):
        return a - b

    def nop(a):
        return a

    functions = [convnum, brac, div, nop, mul, nop, add, nop, sub, nop]
    input_str = "5+1-3*4/2"
    true_result = eval(input_str)
    inputbuf = io.StringIO(input_str)
    p = parser.Parser(math, functions)
    assert abs(p.parse(inputbuf) - true_result) <= 1e-6


def test_palindrome():
    palindrome = grammar.Grammar([('S', ['"a"', 'S', '"a"']),
                                  ('S', ['"b"', 'S', '"b"']), ('S', ['"c"'])])

    def half(a, S=None, b=None):
        if b:
            return a + S
        return ""

    functions = [half] * 3
    input_str = "bbacabb"
    inputbuf = io.StringIO(input_str)

    p = parser.Parser(palindrome, functions)
    assert p.parse(inputbuf) == "bba"


@pytest.mark.xfail(strict=True)
def test_palindrome_invalid():
    palindrome = grammar.Grammar([('S', ['"a"', 'S', '"a"']),
                                  ('S', ['"b"', 'S', '"b"']), ('S', ['"c"'])])

    def half(a, S=None, b=None):
        if b:
            return a + S
        return ""

    functions = [half] * 3
    input_str = "abacabb"
    inputbuf = io.StringIO(input_str)

    p = parser.Parser(palindrome, functions)
    assert p.parse(inputbuf) == "aba"


@pytest.mark.xfail(strict=True, raises=EOFError)
def test_eof_error():
    palindrome = grammar.Grammar([('S', ['"a"', 'S', '"a"']),
                                  ('S', ['"b"', 'S', '"b"']), ('S', ['"c"'])])

    def nop(*args):
        pass

    functions = [nop] * 3
    input_str = "abacab"
    inputbuf = io.StringIO(input_str)

    p = parser.Parser(palindrome, functions)
    assert p.parse(inputbuf) == "aba"


def test_eps_grammar():
    g = grammar.Grammar([('a', ['b', 'c']), ('b', []), ('b', ['"b"']),
                         ('c', ['c', '"c"']), ('c', ['"c"'])])

    def reducer(*args):
        return list(args)

    p = parser.Parser(g, [reducer] * len(g))
    assert p.parse(io.StringIO("ccc")) == [[], [[['c'], 'c'], 'c']]



# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only

import pytest
import pathlib
import io
from pypargen.grm import parser


@pytest.mark.slow
def test_bootstrap():
    grm_file = pathlib.Path(__file__).parent / "grammar.grm"
    grm_parser = parser.GrmParser()
    with grm_file.open() as fp:
        grm1 = grm_parser.parse(fp)
    # Grammar grammar is parsed to get the grammar for the grammar

    grm2 = grm_parser.parse(io.StringIO(str(grm1)))

    grm3 = grm_parser.parse(io.StringIO(str(grm2)))

    assert grm1 == grm2 == grm3
    with open(grm_file) as grm_fp:
        assert grm_fp.read() == str(grm3)



# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only


class NFANode(dict[str, set["NFANode"]]):
    _next_id = 1

    def __new__(cls, token=None, *args, **kwargs):
        node = super().__new__(cls, *args, **kwargs)
        node._id = cls._next_id
        cls._next_id += 1
        node.token = token
        return node

    def __init__(self, *args, **kwargs):
        pass

    def add_transition(self, char: str, nxt: "NFANode"):
        assert len(char) <= 1,\
                "Transition can only be added with single or less character"
        if char in self:
            self[char].add(nxt)
            return
        self[char] = {nxt}

    def add_transitions(self, char: str, nxts: list["NFANode"]):
        assert len(char) <= 1,\
                "Transition can only be added with single or less character"
        if char in self:
            self[char].update(nxts)
            return
        self[char] = set(nxts)

    def add_chars_transition(self, chars: str, nxt: "NFANode"):
        for char in chars:
            if char in self:
                self[char].add(nxt)
                continue
            self[char] = {nxt}

    def __eq__(self, other: "NFANode"):
        return self._id == other._id

    def __ne__(self, other: "NFANode"):
        return self._id != other._id

    def __hash__(self):
        return self._id


class NFA:
    def __init__(self, start: NFANode = None, end: NFANode = None):
        if start is None:
            start = NFANode()
        if end is None:
            end = NFANode()
        self.start = start
        self.end = end


class DFANode(set[NFANode]):

    def __init__(self, nfaNodes: set[NFANode]):
        super().__init__(nfaNodes)
        # Calculate the ϵ-closure of nfaNodes
        stack = list(nfaNodes)
        while stack:
            t = stack.pop(0)
            if '' not in t:
                continue
            for state in t['']:
                if state not in self:
                    self.add(state)
                    stack.insert(0, state)

    def move(self, to: str) -> "DFANode":
        destNodes = set()
        for node in self:
            if to in node:
                destNodes.update(node[to])
        return DFANode(destNodes)

    @property
    def tokens(self) -> set[str]:
        tokens = {x.token for x in self if x.token}
        if not tokens:
            return False
        return tokens


class DFA:
    def __init__(self, nfa: NFA):
        self.start = DFANode({nfa.start})

    def match(self, string: str) -> tuple[int, str]:
        state = self.start
        i = 0
        for i, c in enumerate(string):
            nstate = state.move(c)
            if not nstate:
                break
            state = nstate
        else:
            i += 1
        if len(string) == 0:
            i = 0
        if state.tokens:
            return (state.tokens, i)
        return False



# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only

from pypargen.lexer import re


def test_re_star():
    compiled = re.compile('a*b')
    assert compiled.match("abc")[1] == 2
    assert compiled.match("aaabcd")[1] == 4
    assert compiled.match("bbc")[1] == 1
    assert not compiled.match("caaab")


def test_re_or():
    compiled = re.compile('a*b|b*a')
    assert compiled.match("abc")[1] == 2
    assert compiled.match("aaabcd")[1] == 4
    assert not compiled.match("caaab")

    assert compiled.match("bbac")[1] == 3
    assert not compiled.match("bbbbb")


def test_re_rdb():
    compiled = re.compile('(abc)*bca')

    assert compiled.match("abcbca")[1] == 6
    assert compiled.match("abcabcbca")[1] == 9
    assert not compiled.match("abbca")


def test_re_sqb():
    compiled = re.compile('s[a-cF-I%$7-9]*')

    assert compiled.match("saI7%")[1] == 5
    assert compiled.match("s$9cIP")[1] == 5
    assert not compiled.match("$97IP")


def test_re_or_empty():
    compiled = re.compile('[0-9][0-9]*(()|.[0-9][0-9]*)')

    assert compiled.match("0.909")[1] == 5
    assert compiled.match("90")[1] == 2
    assert not compiled.match(".29")


def test_esc_chars():
    compiled = re.compile(r'\[*(\r|\n|\t)*\]*')

    assert compiled.match("[[[")[1] == 3
    assert compiled.match("[[[\n\t]]]")[1] == 8
    assert compiled.match("[[[\r\n\r\n]]")[1] == 9
    assert compiled.match("\n\t]]")[1] == 4
    assert compiled.match("s\n\t]]")[1] == 0
    assert compiled.match("")[1] == 0



# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only

"""Base module containing abstract classes, basic utilities etc."""

from pypargen.base.token import Token
from pypargen.base.rule import Rule

__all__ = ["Token", "Rule"]



# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only

from typing import Iterable, Optional
import io

from pypargen.base.token import Token


class UnexpectedCharacter(Exception):
    """Exception thrown when invalid character is seen."""

    def __init__(self,
                 char: str,
                 pos: int,
                 expected: Optional[list[str]] = None):
        """Initalize the exception with the invalid character and character
        position"""
        self.pos = pos
        self.char = char
        self.expected = expected

        msg = f"Invalid character '{self.char}' at position {self.pos}"
        if expected is not None:
            msg += f"\nExpected one of: {self.expected}"

        super().__init__(msg)


class UnregisteredTerminal(Exception):
    """Exception thrown when an active terminal passed is unregistered"""
    def __init__(self, terminal: str):
        """Initialize the exception with that invalid terminal"""
        self.terminal = terminal
        super().__init__(f"Unregistered terminal: {self.terminal}")


class BaseLexer:
    """BaseLexer is an abstract class for all the lexers
    The lexer API would be used as:
    ```
    terminals = ['"[a-z]+"', '"[A-Z]+"']
    lexer = Lexer(terminals, sys.stdin)
    tokens = list(lexer)
    pos = lexer.pos
    ```
    """
    def __init__(self,
                 terminals: list[str],
                 inpt: io.RawIOBase,
                 whitespaces: Optional[str] = None):
        """Initialize lexer with terminals that should be looked for and input
        stream. The order of terminals may imply precedence."""
        assert all([x.startswith('"') for x in terminals]), \
            "All terminals must start with a \""

        # Make a copy of terminals so that original changes are not reflected
        self.terminals = terminals.copy()
        self.input = inpt
        self.whitespaces = whitespaces

    def nextToken(self, terminals: Optional[list[str]] = None) -> Token:
        """Override the nextToken method based on the lexer

        The parser can call this method to change the terminals to look for.
        The order of terminals may imply precendence."""
        raise NotImplementedError("Use a subclass of BaseLexer")

    def __iter__(self) -> Iterable[Token]:
        return self

    def __next__(self) -> Token:
        """Implements iterator protocol

        This will simply call nextToken method without terminals argument
        and returns the value. Useful for simple grammars.
        """
        return self.nextToken()



# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only

"""This file contains parser and DFA constructor for regular expressions.
The goal of this module is not to compete with Python RE, but provide a
performant regular expression engine for Lexer purposes, though this module
can be used for regular expressions compilation and matching
"""

import io

from pypargen.base.lexer import BaseLexer
from pypargen.grm.grammar import rules
from pypargen.lr1.parser import Grammar, Parser
from pypargen.lexer import fsm, pyre


def nop(a: any):
    return a


callbacks = [nop]


def rng(chr_rng: str) -> str:
    frm, _, to = chr_rng
    assert frm <= to, "Range item invalid"
    return ''.join(map(chr, range(ord(frm), ord(to) + 1)))


callbacks += [rng] * 3


def char(char: str, esc_char: str = None) -> str:
    if char == '\\':
        char = esc_char
    if esc_char == 'r':
        char = '\r'
    if esc_char == 'n':
        char = '\n'
    if esc_char == 't':
        char = '\t'
    return char


callbacks += [char] * 2

# sqc -> rng | chr
callbacks += [nop] * 2


def sqs(sqs: str, sqc: str = ''):
    return sqs + sqc


callbacks += [sqs] * 2


def sq(_left, sqs, _right) -> fsm.NFA:
    nfa = fsm.NFA()
    for c in sqs:
        nfa.start.add_transition(c, nfa.end)
    return nfa


callbacks += [sq]


def rd(_left: str, re: fsm.NFA, _right: str):
    return re


def rd_empty(_left: str, _right: str) -> fsm.NFA:
    empty = fsm.NFA()
    empty.start.add_transition('', empty.end)
    return empty


callbacks += [rd, rd_empty]

# stc -> sq | rd
callbacks += [nop] * 2


# stc -> chr
def char_to_nfa(char: str) -> fsm.NFA:
    nfa = fsm.NFA()
    nfa.start.add_transition(char, nfa.end)
    return nfa


callbacks += [char_to_nfa]


def st(stc: fsm.NFA, _star) -> fsm.NFA:
    nfa = fsm.NFA()
    nfa.start.add_transition('', stc.start)
    stc.end.add_transition('', nfa.end)
    nfa.start.add_transition('', nfa.end)
    stc.end.add_transition('', stc.start)
    return nfa


callbacks += [st]

# rec -> sq | rd | st
callbacks += [nop] * 3

# rec -> chr
callbacks += [char_to_nfa]


def res_append(res: fsm.NFA, rec: fsm.NFA) -> fsm.NFA:
    res.end.add_transition('', rec.start)
    return fsm.NFA(res.start, rec.end)


# res -> res rec | recc
callbacks += [res_append, nop]


def re(re: fsm.NFA, _or: str, res: fsm.NFA) -> fsm.NFA:
    nfa = fsm.NFA()
    nfa.start.add_transition('', re.start)
    nfa.start.add_transition('', res.start)
    re.end.add_transition('', nfa.end)
    res.end.add_transition('', nfa.end)
    return nfa


# re -> re "\|" res | res
callbacks += [re, nop]


re_rules = rules[:len(callbacks)]
re_grm = Grammar(re_rules, "re")


class REParser(Parser):
    """Regular Expression parser. This parser supports only basic
    meta-characters: |, *, (, ), [, ]
    Other meta-characters like ., +, ? are taken literally.

    It is simpler to implement these meta-characters with supported ones:
    (ab)+ -> ab(ab)*
    [0-9]*(.[0-9][0-9]*)? -> [0-9]*((.[0-9][0-9]*)|())
    . -> [ !#$%&'+,-./0-9:;<=>?@A-Z^_`a-z{}~]

    The regular expression must never be an empty string, to indicate empty
    pattern, use "()".
    """

    def __init__(self, lexerClass: type[BaseLexer] = pyre.PyRELexer):
        "Initialize the parser. See help(BaseParser) for more details"
        super().__init__(re_grm, callbacks, lexerClass)

    def parse(self, re: str) -> fsm.NFA:
        "Parse the regular expression from the string"
        return super().parse(io.StringIO(re))


compiler = None


def compile(re: str) -> fsm.DFA:
    """Compile a regular expression into a DFA. The compiled DFA provides
    match method to match a string:
    ```
    dfa = re.compile("a*b")
    dfa.match("aaabb") # returns ({"match"}, 4)
    ```
    """
    global compiler
    if not compiler:
        compiler = REParser()
    nfa = compiler.parse(re)
    nfa.end.token = "match"
    return fsm.DFA(nfa)



# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only

"""This example shows how to build a simple desk calculator with parser
for basic math operations. It supports addition, subtraction, multiplication,
division and brackets, with proper precedence."""

import pypargen as pgen
import sys

math_rules = [("atom", ['"[1-9][0-9]*"']), ("atom", [r'"\("', "sub", r'"\)"']),
              ("div", ["div", '"/"', "atom"]), ("div", ["atom"]),
              ("mul", ["mul", r'"\*"', "div"]), ("mul", ["div"]),
              ("add", ["add", r'"+"', "mul"]), ("add", ["mul"]),
              ("sub", ["sub", '"-"', "add"]), ("sub", ["add"]),
              ("expr", ["sub", r'"\n\n*"'])]

math = pgen.Grammar(math_rules, "expr")


def convnum(a):
    return int(a)


def bracket(a, b, c):
    print(a, b, c)
    return b


def div(a, _, b):
    return a / b


def mul(a, _, b):
    return a * b


def add(a, _, b):
    return a + b


def sub(a, _, b):
    return a - b


def nop(a, *_):
    return a


callbacks = [convnum, bracket, div, nop, mul, nop, add, nop, sub, nop, nop]
parser = pgen.Parser(math, callbacks, pgen.Lexer, whitespaces=" \t")

result = parser.parse(sys.stdin)

print(result)



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['pypargen', 'pypargen.base', 'pypargen.grm', 'pypargen.lexer', 'pypargen.lr1']

package_data = \
{'': ['*']}

extras_require = \
{'test': ['pytest', 'pytest-cov']}

setup(name='pypargen',
      version='0.3.0',
      description='A parser generator built with Python',
      author='Ilango Rajagopal',
      author_email='ilangokul@gmail.com',
      url='https://github.com/ilango100/pypargen',
      packages=packages,
      package_data=package_data,
      extras_require=extras_require,
      python_requires='>=3.9',
     )



# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only

from pypargen.base import rule


def test_str_rule():
    r = rule.Rule('a', ['b', 'c'])
    assert str(r) == "a\t-> b c"


def test_str_eps():
    r = rule.Rule('a', [])
    assert str(r) == "a\t-> ϵ"


def test_repr_rule():
    r = rule.Rule('a', ['b', 'c'])
    assert repr(r) == "<a\t-> b c>"


def test_repr_eps():
    r = rule.Rule('a', [])
    assert repr(r) == "<a\t-> ϵ>"


def test_hash():
    r1 = rule.Rule('a', ['b', 'c'])
    r2 = rule.Rule('a', [])
    assert hash(r1) != hash(r2)



# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only

from typing import Optional
import re
import io

from pypargen.base.token import Token
from pypargen.base.lexer import BaseLexer, UnexpectedCharacter,\
        UnregisteredTerminal


class PyRELexer(BaseLexer):
    """Lexical tokenizer based on re library."""

    def __init__(self,
                 terminals: list[str],
                 inpt: io.RawIOBase,
                 whitespaces: Optional[str] = None):
        """Initialize lexer with terminals to be looked for and input stream"""
        super().__init__(terminals, inpt, whitespaces)
        self._patterns = {patt: re.compile(patt[1:-1]) for patt in terminals}

        self.ws_pattern = re.compile("")
        if self.whitespaces:
            self.ws_pattern = re.compile(f"[{self.whitespaces}]*")

        # Read the whole input (No other way to use re)
        self.str = inpt.read()
        if hasattr(self.str, "decode"):
            self.str = self.str.decode()
        self.pos = 0
        self.stopped = False

    def nextToken(self, terminals: Optional[list[str]] = None) -> Token:
        if self.stopped:
            raise StopIteration

        # First, skip whitespaces
        if ws := self.ws_pattern.match(self.str, self.pos):
            self.pos = ws.end()

        # Generate the last token as $
        if self.pos >= len(self.str):
            self.stopped = True
            return Token('$', None)

        # Check for active patterns
        if not terminals:
            terminals = self.terminals

        # Passing terminals changes "active" terminals to look for
        for patt in terminals:
            if patt not in self.terminals:
                raise UnregisteredTerminal(patt)
            if match := self._patterns[patt].match(self.str, self.pos):
                term = match.group(0)
                self.pos += len(term)
                return Token(patt, term)
        raise UnexpectedCharacter(self.str[self.pos], self.pos, terminals)



# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only

import io
from typing import Optional

from pypargen.base.grammar import BaseGrammar
from pypargen.base.lexer import BaseLexer


class BaseParser:
    """BaseParser is an abstract class for all the parser
    The parser API would be used as:
    ```
    grammar = Grammar(...rules...)
    parser = Parser(grammar)
    parse_tree = parser.parse(sys.stdin)
    ```
    """

    def __init__(self,
                 grammar: BaseGrammar,
                 lexerClass: type[BaseLexer] = BaseLexer,
                 whitespaces: Optional[str] = None):
        "Initialize parser with grammar rules and an optional Lexer Class"
        self.grammar = grammar
        self.lexerClass = lexerClass
        self.whitespaces = whitespaces

    def parse(self, inpt: io.RawIOBase) -> any:
        "Override the parse method based on the parser"
        raise NotImplementedError("Use a subclass of BaseParser")
